c...Plane strain elasto-plasticity "main engine"
c...for use with FELIPE version 3 - sample datafile is plcyl4.dat
c...pladv is same as plast, but includes a range of solution methods
c
c
c...If you want to re-compile this program, you must link it with
c...the file felsb.obj, compiled from felsb.for (provided in the
c...FELIPE folder). The file felsb.for contains additional subroutines 
c...used by all the 'main engines' except eladv.for. 
c...If you don't use linking, you should paste the contents of 
c...the file felsb.for to the end of this file, before compiling
c...and running it.
c
      program pladv
c...pladv uses row-based skyline storage of the global stiffness matrix
c...and allows solution by: 
c... 0)                 T      Choleski decomposition
c... 1)            L.D.L   decomposition
c... 2)  conjugate gradients with diagonal preconditioning
c... 3)  conjugate gradients with Incomplete Choleski preconditioning
c
cc...This program is for advanced elasticity analyses (ndofn=-2) using
cc...the theory of plasticity. It uses Mohr-Coulomb yield function, and
cc...is written for associated flow rule.
cc...Plane strain analysis, with material thickness = 1.0 
c
CC...THE PROGRAM (AND FELSUB.FOR) SHOULD BE COMPILED IN DOUBLE PRECISION:
CC...     ftn77  /dreal plast.for     in Salford Fortran77
c
      parameter(melem=900,mpoin=2000,mnode=8,
     1msets=9,mcomp=6,mtotv=4000,mstif=150000)
c...these parameters fix the dimensions of the arrays:
c...melem = max. no. of elements
c...mpoin = max. no. of nodes
c...mnode = max. no. of nodes in a single element
c...msets = max. no. of mat. prop. sets (also used for loading sets)
c...mcomp = max. no. of mat. props = 6 for (visco)plasticity
c...mtotv = max no. of degrees of freedom (= 2*mpoin for elasticity)
c...mstif = max. length of vector gstif storing global matrix using
c                                               skyline storage
c
      character char*1
      character*80 title,infile,outfil,prtfil
c
      dimension lnods(melem,mnode),lnodn(melem),lsidn(melem),
     1 coord(mpoin,2),nkode(mpoin),aslod(mtotv),ndof(mpoin),
     2        lpros(melem),props(msets,mcomp),asdis(mtotv),
     3      surft(msets,2),gstif(mstif),lmtyp(melem),
     4      nminvc(mtotv),lpoint(mtotv),astmp(mtotv),diagk(mtotv)
     5      ,z(mtotv),r(mtotv),p(mtotv),y(mtotv)
c...lnods(l,n) = global no. of n'th node of element l
c...lnodn(l) = no. of nodes in element l ( = 8 for 8-noded quad.)
c...lsidn = no. of sides in element l    ( = 4 for 8-noded quad.)
c...coord(np,1), coord(np,2) = x,y coords of node np
c...nkode(np) = fixity code of node np
c...aslod is global nodal load vector
c...astmp is a scratch vector used in the equation solution substitutions
c...ndof(np) = no. of dof.s at node np (not used in this program)
c...lpros(l) = mat. prop. set no. of element l
c...props(m,i) = value of i'th mat. property in mat. prop. set m
c...asdis is global nodal displacement vector
c...surft(m,1),surft(m,2) = normal,tangential surface tractions in set m
c...gstif holds assembled global stiffness matrix in compact form
c...lmtyp(l) = material type of elt l (not used in this program)
c...nminvc(i)=j means that G(i,j) is the first nonzero entry in row i of gstif
c...lpoint(i)=ii means that G(i,i) is stored at gstif(ii) in skyline storage
c 
      dimension strsg(3,4,melem),strsg1(3,4,melem),totld(mtotv),
     1          accld(mtotv),accds(mtotv),aslod1(mtotv),resld(mtotv),
     2          fincs(9),mohrg(4,melem)
cc..strsg holds stresses at Gauss points (Cartesian)
cc..totld is the total load to be applied
cc..accld is the accumulated load applied so far
cc..aslod is the load being applied in the current increment
cc..accds is the accumulated displacements 
cc..resld is the residual force vector
cc..fincs is the set of load increments
cc..mohrg is the yield code at Gauss pts: 0=elastic, 1=plastic
c
cc..tolerances for convergence tests:
      gtoler = 1.e-4
      ftoler = 1.e-6
      stoler = 1.e-8
cc..gtoler is tolerance for the global (displacement) iterations
cc..ftoler is tolerance for the yield function F on yield surface
cc..stoler is tolerance for stresses in local (stress) iterations
cc..ftoler and stoler are used in subroutine stress
c
c
      write(6,*)'Program PLADV: advanced plasticity (Mohr-Coulomb).'
      write(6,*)'Handles 8-noded quadrilateral elements.'
      write(6,*)'Solution by Choleski, LDLt or conjugate gradients'
      write(6,*)'with diagonal or Incomplete Cholesk preconditioning.'
      write(6,*)'Dimensioned for',melem,' elements, and',mpoin,' nodes.'
      write(6,*)
c
c
c...read an existing .dat file
c
    1 continue
c
c...get the input file name in infile
      call askfil(infile,iflag)
      if (iflag.eq.1) then
        call yesno(char,'Do you want to exit?')
        if (char.eq.'y') stop
        goto 1
      endif
c
      write(6,*)'Opening file ',infile,'...'
      lenin = lenstr(infile)
      open(15,file=infile(1:lenin),status='old',err=999)
      write(6,*)
      lm4 = lenin - 4
      prtfil = infile(1:lm4)//'.prt'
      write(6,*)'Creating file ',prtfil(1:lenin),' for results...'
      write(6,*)
      open(16,file=prtfil(1:lenin),status='unknown',err=999)      
      outfil = infile(1:lm4)//'.out'
      write(6,*)'Creating file ',outfil(1:lenin),' for output...'
      write(6,*)
      open(17,file=outfil(1:lenin),status='unknown',err=999)      
c
c...read the input mesh data, and echo it to the output files
      call input(15,melem,mpoin,mnode,msets,mcomp,nelem,
     1  npoin,lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props
     2         ,ndofn,npros,title,ndof)
      call print(16,melem,mpoin,mnode,msets,mcomp,nelem,npoin,
     1  lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props,ndofn,
     2           npros,title,ndof)
      call output(17,melem,mpoin,mnode,msets,mcomp,nelem,npoin,
     1  lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props,ndofn,
     2            npros,title,ndof)
      call yesno(char,'Do you want the mesh details listed?')
      if (char.eq.'y') then
        call print(6,melem,mpoin,mnode,msets,mcomp,nelem,npoin,
     1  lnods,lnodn,lsidn,lmtyp,coord,nkode,lpros,props,ndofn,
     2           npros,title,ndof)
      endif
c
      iflag = 0
      do 3 l=1,nelem
        nnode = lnodn(l)
        nside = lsidn(l)
        nodsid = 10*nnode + nside
        if (nodsid.ne.84) then
          iflag = 1
          write(6,*)'ERROR: Cannot handle element ',l
          write(6,*)'which has ',nside,' sides and ',nnode,' nodes'
        endif
    3 continue
      if (iflag.eq.1) then
        call prompt
        stop
      endif
c
      write(6,*)'You can use a direct or an iterative solution method.'
      write(6,*)'Choose the parameter ISOLA from the following:'
      write(6,*)
      write(6,*)'ISOLA = 0 for direct solution using Choleski (LLt)'
      write(6,*)'           decomposition and skyline storage;'
      write(6,*)'ISOLA = 1 for direct solution using L.D.Lt'
      write(6,*)'           decomposition and skyline storage;'
      write(6,*)'ISOLA = 2 for iterative solution using conjugate'
      write(6,*)'           gradients with diagonal preconditioning;'
      write(6,*)'ISOLA = 3 for Incomplete Choleski preconditioning'
      write(6,*)'           with conjugate gradient solution.'
      write(6,*)
      write(6,*)'Type choice of ISOLA now:'
      read(5,*) isola 
      if (isola.gt.1) then
      write(6,*)'Enter tolerance for convergence of c.g. iterations:'
        read(5,*) cgtol
        if (isola.eq.3) then
      write(6,*)'Enter factor for determining if fill-in will occur:'
          read(5,*) factor
        endif
      endif
c
cc..zero the accumulating variables, stresses, etc.
      ntotv = 2*npoin
c
      do 10 i=1,mtotv
        accds(i) = 0.0
        accld(i) = 0.0
        resld(i) = 0.0
   10 continue
      do 11 i=1,9
        fincs(i) = 0.0
   11 continue  
      do 12 l=1,melem
        do 12 ij=1,4
          do 12 i=1,3
            strsg(i,ij,l) = 0.0
            strsg1(i,ij,l) = 0.0
   12 continue
      do 13 l=1,melem
        do 13 ij=1,4
          mohrg(ij,l) = 0
   13 continue
cc..read the equivalent load vector, into totld, and then
cc..read the load increment sizes
      call prompt
      write(6,*)
      write(6,*)'Starting subroutine LOAD...'
      call load(melem,mpoin,mnode,msets,mcomp,mtotv,
     1          nelem,npoin,lnods,lnodn,coord,
     2          surft,totld)
c
c...read boundary data header (no boundary set data expected)
      read(15,*)
      read(15,*)nbrad,nbdir
      if (nbrad.gt.0.or.nbdir.gt.0) then
        write(6,*)
     1  'ERROR: boundary set data present - cannot cope!'
      call prompt
        stop
      endif
c...read data defining load increments
      read(15,*) 
      read(15,*) nincs,(fincs(i),i=1,nincs)
c
c
c...open a scratch file to write the element stiffness matrices on
      open(unit=18,status='scratch',form='unformatted')  
c
c
c...analyse the "mask" of the global stiffness matrix, i.e. the range
c...of non-zero values in each row. This information is stored in
c...nminvc: row i starts in column nminvc(i). This is used to construct 
c...lpoint: g(i,i) will be located at gstif(lpoint(i))
      call mask(melem,mpoin,mnode,mtotv,mstif,ntotv,
     1       nelem,npoin,lnods,lnodn,nminvc,lpoint)
c...doing this now, avoids repeating it within assmb at each iteration!
c
c
cc..start of load increment loop
cc..============================
c
      do 200 inc=1,nincs
        finc = fincs(inc)
        write(6,*)' Load increment ',inc
        write(6,*)
        write(6,*)' Increment size: ',finc
        write(6,*)
cc..increment load, adding residual from previous increment
        do 20 i=1,ntotv
          asdis(i) = 0.0
          aslod(i) = finc*totld(i)
          accld(i) = accld(i) + aslod(i)
          aslod(i) = aslod(i) + resld(i)
   20   continue
        do 21 l=1,nelem
          do 21 ij=1,4
            do 21 i=1,3
              strsg1(i,ij,l) = strsg(i,ij,l)
   21   continue
cc..strsg will stay fixed as the start-of-increment stress
cc..strsg1 will hold the iterates for the end-of-increment stress 
c
        kount = 0
cc..start of global iteration for displacements
cc..===========================================
  100   continue
          kount = kount + 1
cc..note: stiffness depends on current stress state
          call stiff(melem,mpoin,mnode,msets,mcomp,
     1           nelem,npoin,lnods,lnodn,coord,lpros,props
     2           ,strsg1,mohrg)
c
c...don't need to assemble gstif if using PCG with diagonal precond.
          if (isola.ne.2) then
            call assmb(melem,mpoin,mnode,mtotv,mstif,ntotv,
     1                 nelem,npoin,lnods,lnodn,
     2                 gstif,nminvc,lpoint)
          endif
c
c...store the diagonal of K: used as preconditioner if isola=2,
c...and used to calculate reactions
          call assdk(melem,mpoin,mnode,mtotv,
     1                 nelem,npoin,lnods,lnodn,nkode,diagk)
c
cc..test residual, to see if iteration has converged
          call check(mtotv,melem,mpoin,mnode,
     1     ntotv,nelem,npoin,nkode,lnods,lnodn,
     1           asdis,aslod,resld,rnorm,nminvc,lpoint)
          write(6,*)'            Residual norm = ',rnorm
          write(6,*)
          if (kount.ge.2.and.rnorm.lt.gtoler) goto 150
c
c 
          do 110 i=1,ntotv
            aslod1(i) = aslod(i)
  110     continue
c
          if (isola.le.1) then
c...direct solution using Choleski decomposition
            call solve(mpoin,mtotv,mstif,npoin,nkode,
     1        aslod1,asdis,astmp,gstif,nminvc,lpoint,isola)
          else
c...iterative solution using preconditioned conjugate gradients
            call pcgsol(mtotv,melem,mpoin,mnode,ntotv,nelem,
     1      npoin,lnods,lnodn,diagk,nkode,mstif,gstif,nminvc,lpoint,
     2         aslod1,asdis, z , r , p , y , astmp,
     3         cgtol,itscg,rnorm0,isola,factor)
           endif
c
cc..now evaluate new stresses, from this displt increment, in strsg1
cc..and update mohrg if new points have yielded
          call stress(melem,mpoin,mnode,msets,mcomp,mtotv,
     1            nelem,npoin,lnods,lnodn,coord,lpros,props,
     2            asdis,strsg,strsg1,mohrg,ftoler,stoler)
c
          goto 100
cc..end of global iteration
cc..=======================
  150   continue
        write(6,*) kount,' global iterations'
        write(6,*)
c
c...update total displacements and stresses
        do 160 i=1,ntotv
          accds(i) = accds(i) + asdis(i)
  160   continue
        do 161 l=1,nelem
          do 161 ij=1,4
            do 161 i=1,3
              strsg(i,ij,l) = strsg1(i,ij,l)
  161   continue
c
cc..compute residual force vector
        call residu(melem,mpoin,mnode,mtotv,nelem,npoin,ntotv,
     1              lnods,lnodn,coord,nkode,accld,strsg,resld)
c
  200 continue
cc..end of load increment loop
cc..==========================
c
      call result(melem,mpoin,mnode,msets,mcomp,mtotv,
     1            nelem,npoin,lnods,lnodn,coord,nkode,lpros,props,
     2            accld,accds,strsg,mohrg)
c
      write(6,*)
      write(6,*)'PLADV completed. Results are in file ',prtfil(1:lenin)
      write(6,*)
      write(6,*)
      call prompt
      stop
  999 write(6,*)
     1'Cannot open file with this name. Terminating program.'
      call prompt
      stop
      end
c
c
c      
      subroutine stiff(melem,mpoin,mnode,msets,mcomp,
     1             nelem,npoin,lnods,lnodn,coord,lpros,props,
     2             strsg1,mohrg)
c      
c...calculates the element stiffness matrices, and writes them
c...to the scratch file.
c
      dimension props(msets,mcomp),lnods(melem,mnode),encod(2,8),
     1         estif(16,16),bmatx(3,16),dmatx(3,3),dbmat(3,16),
     2        gpwts(4),lpros(melem),lnodn(melem),coord(mpoin,2),
     3       shape(8),gploc(2,4),gpcod(2),deriv(2,8),cartd(2,8)
cc..extra arrays in plasticity
     4  ,strsg1(3,4,melem),gpsig(3),avect(3),bvect(3),davec(3),
     5  dbvec(3),demat(3,3),mohrg(4,melem)
c...encod holds the nodal coords for the nodes around the current element
c...estif will be the stiffness matrix for the current element
c...bmatx, dmatx are the B and D matrices, and dbmat is D*B
c...gpwts,gploc are the Gauss point weights and locations (in xi,eta coords)
c...gpcod will be the x,y coords of the current Gauss point
c...shape holds the shape functions evaluated at the Gauss point
c...deriv holds the derivatives (w.r.t. xi, eta) of the shape functions
c...cartd holds the Cartesian derivatives (w.r.t x,y) of shape functions
c
cc..avect is the gradient of the yield function dF/d(sigma)
cc..bvect is the gradient of the plastic potential dQ/d(sigma)
cc..demat is the elastic stiffness matrix
c
      rewind 18
c
c...start loop over all elements
      do 80 l=1,nelem
c
c...extract Young's modulus & Poisson's ratio from props      
      lset = lpros(l)
      ym = props(lset,1)
      pr = props(lset,2)
cc..and the Mohr-Coulomb plasticity parameters k and sigma_c
      triax = props(lset,3)
      sigc = props(lset,4)
cc...set material thickness to 1.0
      thick = 1.0
c
c...put nodal coords in encod
      nnode = lnodn(l)
      do 15 n=1,nnode
      nn = lnods(l,n)
      encod(1,n) = coord(nn,1)
      encod(2,n) = coord(nn,2)
   15 continue
c...zero stiffness matrix
      do 20 j=1,16
      do 20 i=1,16
      estif(i,j) = 0.0
   20 continue
c
c...get elastic D matrix
      call dmat(ym,pr,demat)
c
c...put local coords of 2x2 guass-pts in gploc
      call gauss(gploc,gpwts,ngaus)
c
c...start loop through gauss-pts for contribution to estif
      do 75 ij=1,ngaus
c...put local coords of gauss-pt in s,t
      s = gploc(1,ij)
      t = gploc(2,ij)
c...get shape fns (shape) & derivs (deriv) in local coords
      call sfr(s,t,shape,deriv)
c...get xy-coords of gauss-pt into gpcod, using shape fn interpolation
      do 25 m=1,2
      sum = 0.0
      do 24 n=1,8
   24 sum = sum + encod(m,n)*shape(n)
   25 gpcod(m) = sum
c...get xy-derivs (cartd) & derterminant of Jacobian (djacb)
      call jacob(encod,shape,deriv,cartd,djacb)
c...get B matrix 
      call bmat(cartd,bmatx)
c
cc..extract stress at Gauss point into gpsig
      do 30 i=1,3
        gpsig(i) = strsg1(i,ij,l)
   30 continue
c
      if (mohrg(ij,l).eq.0) then
cc..still elastic - use demat
        do 32 i=1,3
          do 32 j=1,3
            dmatx(i,j) = demat(i,j)
   32   continue
      else
cc..calculate elasto-plastic D matrix
cc..for associated flow, Q=F so avect=bvect
        call flowv(triax,gpsig,avect)
        call flowv(triax,gpsig,bvect)
        atdb = 0.0
        do 35 i=1,3
          dai = 0.0
          dbi = 0.0
          do 34 j=1,3
            dai = dai + demat(i,j)*avect(j)
            dbi = dbi + demat(i,j)*bvect(j)
   34     continue
          davec(i) = dai
          dbvec(i) = dbi
          atdb = atdb + avect(i)*dbi
   35   continue
        do 37 i=1,3
          do 37 j=1,3
            dmatx(i,j) = demat(i,j) - dbvec(i)*davec(j)/atdb
   37   continue
      endif
cc..end of plasticity section
c
c...form matrix product D*B
      dvolu = thick*djacb*gpwts(ij)
      do 40 j=1,16
      do 40 i=1,3
      sum=0.0
      do 39 k=1,3
   39 sum = sum + dmatx(i,k)*bmatx(k,j)
   40 dbmat(i,j) = sum
c...B transpose *d*b* djacb = contribution to estif
      do 50 j=1,16
      do 50 i=1,16
      sum = 0.0
      do 49 k=1,3
   49 sum = sum + bmatx(k,i)*dbmat(k,j)
   50 estif(i,j) = estif(i,j) + sum*dvolu
   75 continue
c...end of gauss loop
c
c...write estif to file
      write (18) estif
   80 continue
c...end of element loop
c
      endfile 18
c
      return
      end
c
c
      subroutine load(melem,mpoin,mnode,msets,mcomp,mtotv,
     1                nelem,npoin,lnods,lnodn,coord,surft,aslod)
c
c...compile global nodal load vector
c
      dimension aslod(mtotv),surft(msets,2),
     1     lnods(melem,mnode),press(2,3),npals(3),
     2     coord(mpoin,2),gploc(2),shape(3),deriv(3),xjacm(2),      
     3     gpres(2),fcomp(2),encod(2,3),lnodn(melem)
c
      ntotv = npoin*2
c
      write(6,*)'Applied Loads'
      write(6,*)'-------------'
      write(6,*)
      write(16,*)'Applied Loads'
      write(16,*)'-------------'
      write(16,*)
c...initialize the global load vector
      do 10 i=1,mtotv
      aslod(i) = 0.0
   10 continue
c
c...there are some dummy lines in the input data, which are read but ignored
      read(15,*,end=99)
c...npset>0 if there are surface tractions. nbset is not used.
      read (15,510) nbset,npset
  510 format(2i5)
c
c...nodal loads (ld=1, ignored)
      write(6,*)'Nodal loads:'
      write(6,*)'-----------'
      write(6,*)
      write(6,*)'  Node   x-component   y-component'
      write(16,*)'Nodal loads:'
      write(16,*)'-----------'
      write(16,*)
      write(16,*)'  Node   x-component   y-component'
      read(15,*)
   15 read(15,520) lcard,ld,np,fx,fy
  520 format(i1,i4,i5,2e10.3)
      if (np.ne.0) then
c...the i'th component of node np must be converted to global degree 
c...of freedom component
        nm = 2*(np - 1)
c...add nodal loads into aslod
        aslod(nm + 1) = aslod(nm + 1) + fx
        aslod(nm + 2) = aslod(nm + 2) + fy
        write(6,610) np,fx,fy
        write(16,610) np,fx,fy
  610   format(i5,2(4x,e10.3))
        if (lcard.eq.0) goto 15
      endif
      write(6,*)
      write(16,*)
c
c
c...surface tractions
      if (npset.gt.0) then
        read(15,*)
        write(6,*)'Surface Tractions'
        write(6,*)'-----------------'
        write(6,*)
        write(6,*)' set   normal cmpnt   tangential cmpnt' 
        write(16,*)'Surface Tractions'
        write(16,*)'-----------------'
        write(16,*)
        write(16,*)' set   normal cmpnt   tangential cmpnt' 
c
c...first read surface traction sets
   25   read(15,530) lcard,iset,i1,i2,pn,pt
  530   format(i1,i4,3x,2i2,8x,2e10.3)
        if (iset.gt.0) then 
          surft(iset,1) = pn
          surft(iset,2) = pt
          write(6,620)  iset,pn,pt
          write(16,620) iset,pn,pt
  620     format(i3,2(5x,e10.3))
        endif
        if (lcard.eq.0) goto 25
c
c...now apply these tractions to edges
        read(15,*)
        write(6,*)
        write(6,*)'   elt  end-nodes   set-numbers'
        write(16,*)
        write(16,*)'   elt   end-nodes   set-numbers'
   30   read(15,540) lcard,nnas,l,n1,iset1,n2,iset2
c...nnas is the number of nodes along the edge (=3 for 8-noded quad.)
  540   format(i1,i4,i5,2(i5,i2,1x))
        if (l.gt.0) then
          write(6,630) l,n1,n2,iset1,iset2
          write(16,630) l,n1,n2,iset1,iset2
  630     format(i5,2(2x,2i5))
c...put the tractions in press. Interpolate linearly for the
c...tractions at the midside node.
          do 35 i=1,2
            press(i,1) = surft(iset1,i)
            press(i,3) = surft(iset2,i)
            press(i,2) = (press(i,1) + press(i,3))/2.
   35     continue
c...store the node numbers along side in npals
          npals(1) = n1
          npals(3) = n2
c...we have to find the midside node no. from lnods
          nnode = lnodn(l)
          do 42 n=1,nnode
            if (lnods(l,n).eq.npals(1)) goto 45
   42     continue
c...if loop completed, npals(1) is not in element 1
          write(6,700) npals(1),l
  700     format('ERROR: node',i4,' is not in element ',i4,'.')
          call prompt
          stop
   45     k = n + 1
          npals(2) = lnods(l,k)
c
c...extract nodal coords and store in encod
          do 60 n=1,3
            np = npals(n)
            do 55 m=1,2
   55         encod(m,n) = coord(np,m)
   60     continue
c
c...enter 1D Gauss loop along side (2-pt integration)
          r13 = 1.0/sqrt(3.0)
          gploc(1) = -r13
          gploc(2) = r13
          gpwts = 1.0
          do 90 ij=1,2
            s = gploc(ij)
            ss = s*s
            shape(1) = (ss-s)*0.5
            shape(2) = 1.0 - ss
            shape(3) = (ss + s)*0.5
            deriv(1) = s - 0.5
            deriv(2) = -2.0*s
            deriv(3) = s + 0.5
c...get jacobian xjacm
            do 75 j=1,2 
              sum = 0.0
              do 72 n=1,3
   72           sum = sum + deriv(n)*encod(j,n)
              xjacm(j) = sum
   75       continue
c...find gauss pt pressures, & convert to components in local coords
            do 80 m=1,2
              sum = 0.0
              do 77 n=1,3
   77           sum = sum + press(m,n)*shape(n)
              gpres(m) = sum
   80       continue
            fcomp(1) = (xjacm(1)*gpres(2) - xjacm(2)*gpres(1))*gpwts
            fcomp(2) = (xjacm(1)*gpres(1) + xjacm(2)*gpres(2))*gpwts
c...accumulate equivalent nodal forces in aslod
            do 85 n=1,3
              np = npals(n)
              do 82 m=1,2
                nm = 2*(np - 1) + m
                aslod(nm) = aslod(nm) + fcomp(m)*shape(n)
   82         continue
   85       continue
   90     continue
c...end of Gauss loop
c
        endif
        if (lcard.eq.0) goto 30
c...go back for next loaded edge
c
c...report the non-zero loads assembled in aslod
        call prompt
        write(6,*) 
        write(6,*)'Non-zero entries in global load vector:'
        write(6,*)'  node  comp.     value'
        write(6,*)
        write(16,*) 
        write(16,*)'Non-zero entries in global load vector:'
        write(16,*)'  node  comp.     value'
        write(16,*)
        do 95 i=1,ntotv
          ai = aslod(i)
c...zero any tiny loads caused by rounding errors
          if (abs(ai).lt.1.e-5) then
            aslod(i) = 0.0
          else
            n = (i+1)/2
            m = i - 2*(n-1)
            write(6,650) n,m,ai
            write(16,650) n,m,ai
  650       format(i5,i5,5x,e10.3)
          endif
  95    continue
      endif
c
      write(6,*)
      write(6,*)'End of load data.'
      write(6,*)
      write(16,*)
      write(16,*)'End of load data.'
      write(16,*)
      return
   99 write(6,*)'ERROR: this input file has no loading data.'
      call prompt
      stop
      end
c      
c
      subroutine mask(melem,mpoin,mnode,mtotv,mstif,ntotv,
     1       nelem,npoin,lnods,lnodn,nminvc,lpoint)
c
c...determines the "mask" of gstif, i.e. the location of the nonzero 
c...section of each row. This is stored in nminvc and lpoint
c
      dimension lnods(melem,mnode),
     1         lnodn(melem),nminvc(mtotv),lpoint(mtotv)
c
      ntotv = 2*npoin
c...first, a mock run-through of the assembly, to find the location of the
c...first non-zero element on each row i: store this in nminvc(i)
      do 2 i=1,ntotv
        nminvc(i) = i
    2 continue
c...loop through all elements
      do 25 l=1,nelem
        nnode = lnodn(l)
c...deal with elt nodes ni in rows, nj in columns
c... these will be global nodes nni, nnj
c...note this assumes 2 d.o.f.s per node
        ni=0
    5   ni = ni + 1
        i = 2*(ni - 1)
        nni = lnods(l,ni)
        ii = 2*(nni - 1)
        nj = 0
   10   nj = nj + 1
        j = 2*(nj - 1)
        nnj = lnods(l,nj)
        jj = 2*(nnj - 1)
c...now we could put the four entries from estif into gstif
        do 20 mi=1,2
          do 15 mj=1,2
            im = i + mi
            jm = j + mj
            iim = ii + mi
            jjm = jj + mj
            if (jjm.gt.iim) goto 15
c...ignore entries above diagonal in gstif
c...now we know that the (im,jm) entry of element l will
c...add into the (iim,jjm) position of K
            imin = nminvc(iim)
            if (jjm.lt.imin) imin = jjm
            nminvc(iim) = imin
   15     continue
   20   continue
        if (nj.lt.nnode) goto 10
        if (ni.lt.nnode) goto  5
   25 continue
c...end of element loop
c
c...now find halfbandwidth, and fill in lpoint
      nhbw = 0
      idiag = 0
      do 30 i=1,ntotv
        idiff = i - nminvc(i) 
        if (idiff.gt.nhbw) nhbw = idiff
        idiag = idiag + idiff + 1
        lpoint(i) = idiag 
   30 continue
c...in skyline storage, G(i,i) is stored at gstif(lpoint(i))
c
      write(6,*) ' Global matrix half-bandwidth = ', nhbw
      nstif = idiag
      write(6,*)' Length of skyline storage vector gstif is',nstif
      if (nstif.gt.mstif) then
        write(6,*)
     1   'ERROR: not enough storage. Increase parameter mstif to '
     1            ,nstif,' at start of program, and recompile.'
        call prompt
        stop
      endif
c
      return
      end
c
c
      subroutine assmb(melem,mpoin,mnode,mtotv,mstif,ntotv,
     1       nelem,npoin,lnods,lnodn,gstif,nminvc,lpoint)
c
c...assembles the elt stiffness matrices, read from the scratch file,
c...into the global matrix gstif, stored compactly using its skyline
c...envelope structure
c
      dimension lnods(melem,mnode),estif(16,16),gstif(mstif),
     1         lnodn(melem),nminvc(mtotv),lpoint(mtotv)
c
c...initialize gstif
      do 35 j=1,mstif
        gstif(j) = 0.0
   35 continue
c
      rewind 18
c
c...loop through all elements
      do 80 l=1,nelem
        read (18) estif
c
        nnode = lnodn(l)
c...deal with elt nodes ni in rows, nj in columns
c... these will be global nodes nni, nnj
        ni=0
   55   ni = ni + 1
        i = 2*(ni - 1)
        nni = lnods(l,ni)
        ii = 2*(nni - 1)
        nj = 0
   60   nj = nj + 1
        j = 2*(nj - 1)
        nnj = lnods(l,nj)
        jj = 2*(nnj - 1)
c...put the four entries from estif into gstif
        do 70 mi=1,2
          do 65 mj=1,2
            im = i + mi
            jm = j + mj
            iim = ii + mi
            jjm = jj + mj
            if (jjm.gt.iim) goto 65
c...ignore entries above diagonal in gstif
c...translate (iim,jjm) to entry of array gstif with skyline storage
            ijpos = gpos(lpoint,ntotv,iim,jjm)            
            gstif(ijpos) = gstif(ijpos) + estif(im,jm)
   65     continue
   70   continue
        if (nj.lt.nnode) goto 60
        if (ni.lt.nnode) goto 55
   80 continue
      write(6,*)'Global stiffness matrix assembled.'
c...end of element loop
c
      return
      end
c
c
      function gpos(lpoint,ntotv,i,j)
      dimension lpoint(*)
c...returns the position in the array gstif of a(i,j)
      if (i.lt.1.or.j.lt.1.or.i.gt.ntotv.or.j.gt.ntotv) then
        gpos = 0
        return
      endif
c...map entries in upper triangle, into lower triangle
      if (j.le.i) then
        ii = i
        jj = j
      else
        ii = j
        jj = i
      endif
      if (ii.eq.1) then
        gpos = 1
      else
        lpi = lpoint(ii)
        lpi1 = lpoint(ii-1)
        jmin = ii - lpi + lpi1 + 1
        if (jj.ge.jmin) then
          gpos = lpi - ii + jj
        else
          gpos = 0
        endif
      endif
      return
      end
c
c
      subroutine solve(mpoin,mtotv,mstif,npoin,nkode,
     1     aslod,asdis,astmp,gstif,nminvc,lpoint,isola)
      dimension gstif(mstif),aslod(mtotv),asdis(mtotv),
     1    nkode(mpoin),nminvc(mtotv),astmp(mtotv),lpoint(mtotv)
c
c...solve matrix system  gstif*asdis=aslod
c
      ntotv = 2*npoin
c
c...add 10e30 to diagonal elements for fixed dof 
      call fixdof(mtotv,mstif,ntotv,mpoin,npoin,gstif,
     1            nkode,lpoint)
c
      write(6,*)
      write(6,*)'Starting equation solution: decomposition...'
c...solution of equations by Choleski or LDLt decomposition
      if (isola.eq.0) then
c...Choleski decomposition
        call chodec(mtotv,mstif,ntotv,gstif,lpoint)
      else
c        T
c...L D L   decomposition
        call ldldec(mtotv,mstif,ntotv,gstif,lpoint)
      endif
c
      write(6,*)
      write(6,*)'Starting substitution...'
c...substitution routine depends of which decomposition used
      if (isola.eq.0) then
        call chosub(mtotv,mstif,ntotv,gstif,aslod,asdis,astmp,
     1            lpoint)
      else
        call ldlsub(mtotv,mstif,ntotv,gstif,aslod,asdis,astmp,
     1            lpoint)
      endif
c
      return
      end
c
c
      subroutine result(melem,mpoin,mnode,msets,mcomp,mtotv,
     1                  nelem,npoin,lnods,lnodn,coord,nkode,
     3                  lpros,props,aslod,asdis,strsg,mohrg)
c
c...output the results to file
c
      dimension aslod(mtotv),props(msets,mcomp),
     1         lnods(melem,mnode),coord(mpoin,2),encod(2,8),
     2        shape(8),deriv(2,8),gpwts(4),gpsig(3),gpcod(2),
     3       gploc(2,4),strsg(3,4,melem),mohrg(4,melem),
     4      nkode(mpoin),asdis(mtotv),lnodn(melem),lpros(melem)
c...output displacements 
c...output to screen (unit 6):
      write(6,*)
      write(6,*)'RESULTS'
      write(6,*)'======='
      write(6,*)
      write(6,610)
  610 format(//' nodal displacements'
     1/19('-')//3x,'node',
     212x,'x-reaction',5x,'y-reaction',9x,'x-displt',7x,'y-displt')
c...output to results file (unit 16):
      write(16,*)
      write(16,*)'RESULTS'
      write(16,*)'======='
      write(16,*)
      write(16,610)
c...and output to end of data file, for postprocessor:
      write(17,*)'  node  x-displt  y-displt'
      do 20 n=1,npoin  
        i = 2*(n - 1)
        dispx = asdis(i+1)
        dispy = asdis(i+2)
        write(17,'(i5,2e10.3)') n,dispx,dispy
        if (mod(n,25).eq.0) call prompt
        write(6,615) n,aslod(i+1),aslod(i+2),dispx,dispy 
        write(16,615) n,aslod(i+1),aslod(i+2),dispx,dispy 
  615   format(3x,i3,11x,e10.3,5x,e10.3,10x,e10.3,5x,e10.3)    
   20 continue
c
      call prompt
c...element Gauss pt stressess
  620 format(//'Element Gauss pt. stresses:'/27('-')/)
      write(6,620)
      write(16,620)
      write(17,*)
     1' elt gp    x-coord   y-coord   sigma-x   sigma-y   tau-xy'
      do 100 l=1,nelem
        if (mod(l,3).eq.0) call prompt
        write(6,630) l
        write(16,630) l
  630   format(//' element',i5/1x,12('-')/)
        nnode = lnodn(l)
c...store nodal coords for elt in encod
        do 25 n=1,nnode
          nn = lnods(l,n)
          encod(1,n) = coord(nn,1)
          encod(2,n) = coord(nn,2)
   25   continue
c...find Gauss-pt stresses
        write(6,640) (i,i=1,3)
        write(16,640) (i,i=1,3)
  640   format(/' gauss-pt stresses and strains'/6x,'x-coord',8x,
     1  'y-coord',10x,3('stress (',i1,')',5x))
        call gauss(gploc,gpwts,ngaus)
c...loop over each Gauss point
        do 80 ij=1,ngaus
          s = gploc(1,ij)
          t = gploc(2,ij)
          call sfr(s,t,shape,deriv)
c...get Gauss point coords in gpcod
          do 27 m=1,2
            sum = 0.0
            do 26 n=1,8
   26       sum = sum + encod(m,n)*shape(n)
            gpcod(m) = sum
   27     continue
          do 40 i=1,3
            gpsig(i) = strsg(i,ij,l)
   40     continue
cc..print the plasticity code for this Gauss point
          mkode = mohrg(ij,l)
c...only stresses printed; modify this to print strains too
          write(17,'(i5,i2,i3,5e10.3)') 
     1       l,ij,mkode,(gpcod(i),i=1,2),(gpsig(i),i=1,3)
          write(6,650) (gpcod(i),i=1,2),(gpsig(i),i=1,3)
          write(16,650) (gpcod(i),i=1,2),(gpsig(i),i=1,3)
  650     format(3x,2(e10.3,5x),5x,3(e10.3,5x))
          if (mkode.ne.0) then
            write(6,*)'Gauss-point yielded: kode =',mkode
            write(16,*)'Gauss-point yielded: kode =',mkode
          endif
   80   continue
c...end of Gauss loop
  100 continue
c...end of element loop
c
      write(6,*)
      write(6,*)'END OF RESULTS'
      write(6,*)
      write(16,*)
      write(16,*)'END OF RESULTS'
      write(16,*)
      return
      end
c
c
c...subroutines for STIFF     
c
c
      subroutine gauss(gploc,gpwts,ngaus)
      dimension gploc(2,4),gpwts(4)
c...set up Gauss pts & weights for 2x2 integration
      ngaus = 4
      do 5 i=1,4
        gpwts(i) = 1.0
    5 continue
      r13 = 1.0/sqrt(3.0)
      gploc(1,1) = -r13
      gploc(1,2) = -r13
      gploc(1,3) = r13
      gploc(1,4) = r13
      gploc(2,1) = -r13
      gploc(2,2) = r13
      gploc(2,3) = -r13
      gploc(2,4) = r13
      return
      end
c
c
      subroutine dmat(ym,pr,dmatx)
      dimension dmatx(3,3)
c...ym,pr are Young's mod., Poisson's ratio
c...form the D matrix for plane strain
      g = 0.5*ym/(1.0 + pr)
      yl = g*pr/(0.5 - pr)
      g2yl = 2.*g + yl
      dmatx(1,1) = g2yl
      dmatx(1,2) = yl
      dmatx(1,3) = 0.0
      dmatx(2,1) = yl
      dmatx(2,2) = g2yl
      dmatx(2,3) = 0.0
      dmatx(3,1) = 0.0
      dmatx(3,2) = 0.0
      dmatx(3,3) = g
      return
      end
c
c
      subroutine bmat(cartd,bmatx)
      dimension cartd(2,8),bmatx(3,16)
c...form the B matrix
      do 5 j=1,16
      do 5 i=1,3
        bmatx(i,j) = 0.0
    5 continue
      j = 0
      do 10 n=1,8
        i = j + 1
        j = i + 1
        bmatx(1,i) = -cartd(1,n)
        bmatx(2,j) = -cartd(2,n)
        bmatx(3,i) = -cartd(2,n)
        bmatx(3,j) = -cartd(1,n)
c...negative sign for 'compression +ve' sign convention
   10 continue
      return
      end
c
c
      subroutine sfr(s,t,shape,deriv)
      dimension shape(8),deriv(2,8)
c...evaluate the shape functions and 1st derivs at local coords s,t
      s2 = 2.0*s
      t2 = 2.0*t
      ss = s*s
      tt = t*t
      st = s*t
      sst = ss*t
      stt = s*tt
      st2 = 2.0*st
      shape(1) = (-1.0 + st + ss + tt - sst - stt)*0.25
      shape(2) = (1.0 - t - ss + sst)*0.5
      shape(3) = (-1.0 - st + ss + tt - sst + stt)*0.25
      shape(4) = (1.0 + s - tt - stt)*0.5
      shape(5) = (-1.0 + st + ss + tt + sst + stt)*0.25
      shape(6) = (1.0 + t - ss - sst)*0.5
      shape(7) = (-1.0 - st + ss + tt + sst - stt)*0.25
      shape(8) = (1.0 - s - tt + stt)*0.5
c     
      deriv(1,1) = (t + s2 - st2 - tt)*0.25
      deriv(1,2) = -s + st
      deriv(1,3) = (-t + s2 - st2 + tt)*0.25
      deriv(1,4) = (1.0 - tt)*0.5
      deriv(1,5) = (t + s2 + st2 + tt)*0.25
      deriv(1,6) = -s - st
      deriv(1,7) = (-t + s2 + st2 - tt)*0.25
      deriv(1,8) = (-1.0 + tt)*0.5
      deriv(2,1) = (s + t2 - ss - st2)*0.25
      deriv(2,2) = (-1.0 + ss)*0.5
      deriv(2,3) = (-s + t2 - ss + st2)*0.25
      deriv(2,4) = -t - st             
      deriv(2,5) = (s + t2 + ss + st2)*0.25
      deriv(2,6) = (1.0 - ss)*0.5
      deriv(2,7) = (-s + t2 + ss - st2)*0.25
      deriv(2,8) = -t + st
      return
      end
c
c     
      subroutine jacob(encod,shape,deriv,cartd,djacb)
      dimension shape(8),deriv(2,8),cartd(2,8),encod(2,8),xjacm(2,2),
     1         xjaci(2,2)
c...evaluate the Jacobian matrix, and its determinant
      do 5 j=1,2
      do 5 i=1,2
        sum = 0.0
        do 4 n=1,8
    4   sum = sum + deriv(i,n)*encod(j,n)
        xjacm(i,j) = sum
    5 continue
c...xjacm is the Jacobian matrix. find det & inverse by Cramer        
      djacb = xjacm(1,1)*xjacm(2,2) - xjacm(1,2)*xjacm(2,1)
      if (djacb.lt.1.0e-12) then
        write(6,*)'ERROR: determinant of Jacobian is -ve or zero'
        write(6,*)'Element is inside out, or nodes defined clockwise'
        call prompt
        stop
      endif
      xjaci(1,1) = xjacm(2,2)/djacb
      xjaci(2,2) = xjacm(1,1)/djacb
      xjaci(1,2) = -xjacm(1,2)/djacb
      xjaci(2,1) = -xjacm(2,1)/djacb
c...get Cartesian derivs of shape functions using J inverse
      do 50 i=1,2
      do 50 n=1,8
        sum = 0.0
        do 49 j=1,2
   49   sum = sum + xjaci(i,j)*deriv(j,n)
        cartd(i,n) = sum
   50 continue
      return
      end
c
c
c...subroutines for SOLVE
c
c
      subroutine fixdof(mtotv,mstif,ntotv,mpoin,npoin,gstif,
     1                  nkode,lpoint)
      dimension gstif(mstif),kd(2),nkode(mpoin),
     1          lpoint(mtotv)
c...add a large number to the diagonal coefft of fixed d.o.f.s
      do 20 n=1,npoin
        kode = nkode(n)
        kd(1) = kode/10
        kd(2) = kode - kd(1)*10
        do 10 m=1,2
          if (kd(m).eq.0) goto 10
          nm = 2*(n - 1) + m
          nmpos = lpoint(nm)
          gstif(nmpos) = gstif(nmpos) + 1.0e12
   10   continue
   20 continue
      return
      end
c
c
      subroutine chodec(mtotv,mstif,ntotv,gstif,
     1                  lpoint)
      dimension gstif(mstif),lpoint(mtotv)
c...Choleski decomposition, using skyline storage 
c
      small = 1.e-8
c...find the first diagonal entry, i.e. l(1,1)
      i = 1
      a11 = gstif(1)
      if (a11.lt.small) then
        write(6,*)'ERROR: zero or negative pivot in row ',i
        call prompt
        stop
      endif
      cl11 = sqrt(a11)
      gstif(1) = cl11
c...decompose each row in turn
      do 20 i=2,ntotv
        lpi = lpoint(i)
        lpi1 = lpoint(i-1)
        jstart = i - lpi + lpi1 + 1
c...work along the row
        do 10 j=jstart,i-1
          jjpos = lpoint(j)
          cljj = gstif(jjpos)
          ijpos = gpos(lpoint,ntotv,i,j)
          aij = gstif(ijpos)
          sum = 0.0
          lpj = lpoint(j)
          if (j.eq.1) then
            lpj1 = 0
          else
            lpj1 = lpoint(j-1)
          endif
          kstart=max(i-lpi+lpi1+1,j-lpj+lpj1+1,1)
          do 5 k=kstart,j-1
            ikpos = gpos(lpoint,ntotv,i,k)
            jkpos = gpos(lpoint,ntotv,j,k)
            if (ikpos.eq.0.or.jkpos.eq.0) goto 5
            sum = sum + gstif(ikpos)*gstif(jkpos)
    5     continue
          clij = (aij - sum)/cljj
          ijpos = gpos(lpoint,ntotv,i,j)
          gstif(ijpos) = clij
   10   continue
c...finally find the diagonal entry on the row
        iipos = lpoint(i)
        aii = gstif(iipos)
        sum = 0.0
        kstart = i - lpi +lpi1 + 1
        do 15 k=kstart,i-1
          ikpos = gpos(lpoint,ntotv,i,k)
          clik = gstif(ikpos)
          sum = sum + clik*clik
   15   continue
        clii2 = aii - sum
        if (clii2.lt.small) then
          write(6,*)'ERROR: zero or negative pivot in row ',i
          call prompt
          stop
        endif
        clii = sqrt(clii2)
        iipos = lpoint(i)
        gstif(iipos) = clii
   20 continue
      return
      end
c
c
      subroutine chosub(mtotv,mstif,ntotv,gstif,
     1                  asvin,asvout,astmp,lpoint)
      dimension gstif(mstif),asvout(mtotv),asvin(mtotv)
     1         ,lpoint(mtotv),astmp(mtotv)
c
c...solve gstif*asvout = asvin where gstif holds Choleski decomp.
c
c...forward substitution: solve Ly = b
c...store y in astmp
      astmp(1) = asvin(1)/gstif(1)
      do 10 i=2,ntotv
        bi = asvin(i)
        sum = 0.0
        lpi = lpoint(i)
        lpi1 = lpoint(i-1)
        kstart = i - lpi + lpi1 + 1
        do 5 k=kstart,i-1
          ak = astmp(k)
          if (ak.eq.0.0) goto 5
          ikpos = gpos(lpoint,ntotv,i,k)
          sum = sum + gstif(ikpos)*ak
    5   continue
        iipos = lpoint(i)
        clii = gstif(iipos)
        astmp(i) = (bi - sum)/clii
   10 continue
c                                   T
c...back-substitution phase: solve L x = y
c...We can do this in a column-based manner!
      do 20 i=ntotv,1,-1 
        yi = astmp(i)
        lpi = lpoint(i)
        clii = gstif(lpi)
        xi = yi/clii
        asvout(i) = xi
c...once we have found x_i, transfer the term involving this, to
c...the r.h.s. in all earlier equations
        if (i.eq.1) goto 20
        lpi1 = lpoint(i-1)
        ni = lpi - lpi1 - 1
        do 15 k=1,ni
          clij = gstif(lpi-k)
          if (clij.eq.0.0) goto 15
          jpos = i - k
          astmp(jpos) = astmp(jpos) - clij*xi
   15   continue
   20 continue
      return
      end
c
c
c
      subroutine ldldec(mtotv,mstif,ntotv,gstif,
     1                  lpoint)
      dimension gstif(mstif),lpoint(mtotv)
c        T
c...L D L   decomposition, using skyline storage 
cp...compare this with routine chodec in pladv.for,
cp                              T
cp...which uses the Choleski L L   decomposition
c
      small = 1.e-8
c...find the first diagonal entry, i.e. l(1,1)
      i = 1
      a11 = gstif(1)
      if (abs(a11).lt.small) then
        write(6,*)'ERROR: zero pivot in row ',i
        call prompt
        stop
      endif
      gstif(1) = a11
c...decompose each row in turn
      do 20 i=2,ntotv
        lpi = lpoint(i)
        lpi1 = lpoint(i-1)
        jstart = i - lpi + lpi1 + 1
c...work along the row
        do 10 j=jstart,i-1
          jjpos = lpoint(j)
          djj = gstif(jjpos)
          ijpos = gpos(lpoint,ntotv,i,j)
          aij = gstif(ijpos)
          sum = 0.0
          lpj = lpoint(j)
          if (j.eq.1) then
            lpj1 = 0
          else
            lpj1 = lpoint(j-1)
          endif
          kstart=max(i-lpi+lpi1+1,j-lpj+lpj1+1,1)
          do 5 k=kstart,j-1
            kkpos = lpoint(k)
            dkk = gstif(kkpos)
            ikpos = gpos(lpoint,ntotv,i,k)
            jkpos = gpos(lpoint,ntotv,j,k)
            if (ikpos.eq.0.or.jkpos.eq.0) goto 5
            sum = sum + dkk*gstif(ikpos)*gstif(jkpos)
    5     continue
          clij = (aij - sum)/djj
          gstif(ijpos) = clij
   10   continue
c...finally find the diagonal entry on the row
        iipos = lpoint(i)
        aii = gstif(iipos)
        sum = 0.0
        kstart = i - lpi +lpi1 + 1
        do 15 k=kstart,i-1
          kkpos = lpoint(k)
          dkk = gstif(kkpos)
          ikpos = gpos(lpoint,ntotv,i,k)
          clik = gstif(ikpos)
          sum = sum + dkk*clik*clik
   15   continue
        clii = aii - sum
        if (abs(clii).lt.small) then
          write(6,*)'ERROR: zero or negative pivot in row ',i
          call prompt
          stop
        endif
        gstif(iipos) = clii
   20 continue
      return
      end
c
c
      subroutine ldlsub(mtotv,mstif,ntotv,gstif,
     1                  asrhs,asdis,astmp,lpoint)
      dimension gstif(mstif),asdis(mtotv),asrhs(mtotv)
     1         ,lpoint(mtotv),astmp(mtotv)
c
cp...solution of gstif*asdis = asrhs, for the 
c        T
c...L D L   decomposition of gstif, using skyline storage 
c
      small = 1.e-8
c...forward substitution: solve Ly = b
c...store y in astmp
      astmp(1) = asrhs(1)
      do 10 i=2,ntotv
        bi = asrhs(i)
        sum = 0.0
        lpi = lpoint(i)
        lpi1 = lpoint(i-1)
        kstart = i - lpi + lpi1 + 1
        do 5 k=kstart,i-1
          ikpos = gpos(lpoint,ntotv,i,k)
          gik = gstif(ikpos)
          bk = astmp(k)
          sum = sum + gik*bk
    5   continue
        astmp(i) = bi - sum
   10 continue
c             
cp...diagonal bit: solution of Dz = y
      do 12 i=1,ntotv
        lpi = lpoint(i)
        dii = gstif(lpi)
        astmp(i) = astmp(i)/dii
   12 continue
c
c...back-substitution phase: solve L x = z
c...We can do this in a column-based manner!
      do 20 i=ntotv,1,-1 
        xi = astmp(i)
        asdis(i) = xi
        if (abs(xi).lt.small) goto 20
c...once we have found x_i, transfer the term involving this, to
c...the r.h.s. in all earlier equations
        if (i.eq.1) goto 20
        lpi = lpoint(i)
        lpi1 = lpoint(i-1)
        ni = lpi - lpi1 - 1
        do 15 k=1,ni
          clij = gstif(lpi-k)
          jpos = i - k
          astmp(jpos) = astmp(jpos) - clij*xi
   15   continue
   20 continue
c
      return
      end
c
c
c
cc..SUBROUTINES FOR PLASTICITY
c
c
      subroutine stress(melem,mpoin,mnode,msets,mcomp,mtotv,
     1                 nelem,npoin,lnods,lnodn,coord,lpros,props,
     2              asdis,strsg,strsg1,mohrg,ftoler,stoler)
cc..calculates stress increments from asdis, updates strsg into strsg1
cc..and updates yield code in mohrg
c
      dimension props(msets,mcomp),lnods(melem,mnode),encod(2,8),
     1         bmatx(3,16),asdis(mtotv),
     2        gpwts(4),lpros(melem),lnodn(melem),coord(mpoin,2),
     3       shape(8),gploc(2,4),gpcod(2),deriv(2,8),cartd(2,8)
cc..extra arrays in plasticity
     4  ,strsg1(3,4,melem),gpsig(3),gpeps(3),bvect(3),endis(16),
     5  avect(3),dbvec(3),demat(3,3),mohrg(4,melem),strsg(3,4,melem)
     6  ,zmat(3,3),zinv(3,3),gpsig0(3),dbds(3,3),psifn(3)
     7  ,dsigma(3)
c
c...start loop over all elements
      do 80 l=1,nelem
c
c...extract Young's modulus & Poisson's ratio from props      
      lset = lpros(l)
      ym = props(lset,1)
      pr = props(lset,2)
cc..and the Mohr-Coulomb plasticity parameters k and sigma_c
      triax = props(lset,3)
      sigc = props(lset,4)
c
c...put nodal coords in encod
cc..and nodal displacements in endis
      nnode = lnodn(l)
      do 15 n=1,nnode
      k = 2*(n-1)
      nn = lnods(l,n)
      kk = 2*(nn-1)
      endis(k+1) = asdis(kk+1)
      endis(k+2) = asdis(kk+2)
      encod(1,n) = coord(nn,1)
      encod(2,n) = coord(nn,2)
   15 continue
c
c...get elastic D matrix
      call dmat(ym,pr,demat)
c
c...put local coords of 2x2 guass-pts in gploc
      call gauss(gploc,gpwts,ngaus)
c
c...start loop through gauss-pts
      do 75 ij=1,ngaus
c...put local coords of gauss-pt in s,t
      s = gploc(1,ij)
      t = gploc(2,ij)
c...get shape fns (shape) & derivs (deriv) in local coords
      call sfr(s,t,shape,deriv)
c...get xy-coords of gauss-pt into gpcod, using shape fn interpolation
      do 25 m=1,2
      sum = 0.0
      do 24 n=1,8
   24 sum = sum + encod(m,n)*shape(n)
   25 gpcod(m) = sum
c...get xy-derivs (cartd) & derterminant of Jacobian (djacb)
      call jacob(encod,shape,deriv,cartd,djacb)
c...get B matrix 
      call bmat(cartd,bmatx)
c
cc..form (B*displt_inc) in gpeps
      do 33 i=1,3
        gi = 0.0
        do 32 j=1,16
          gi = gi + bmatx(i,j)*endis(j)
   32   continue
        gpeps(i) = gi
   33 continue
c
cc..first guess for new stress
      do 35 i=1,3
        si = 0.0
        do 34 j=1,3
          si = si + demat(i,j)*gpeps(j)
   34   continue
        gpsig(i) = strsg(i,ij,l) + si
        gpsig0(i) = gpsig(i)
   35 continue
c
cc..if Gauss point has already yielded, we have to iterate
      if (mohrg(ij,l).ne.0) goto 39
c
cc..evaluate yield function at trial stress
      call ycrit(triax,sigc,gpsig,fvalu)
      if (fvalu.le.ftoler) goto 70
cc..Gauss point stays elastic - update stress and continue
c
cc..if we reach here, Gauss point has become plastic
cc..mark Gauss point as plastic
      mohrg(ij,l) = 1
c
      write(6,*) l,ij,' plastic yield: f = ',fvalu 
cc..start local stress iteration
c
cc..store the value of lambda in clam
   39 clam = 0.0
      kount = 0
   40 continue
      kount = kount + 1
cc..evaluate F, a, b and db/ds at current stress
      call ycrit(triax,sigc,gpsig,fvalu)
      call flowv(triax,gpsig,avect)
      call flowv(triax,gpsig,bvect)
      call flowm(triax,gpsig,dbds)
cc..form matrix [I + (lambda) D db/d(sigma)] and invert it
cc..also form Db vector
      do 44 i=1,3
        dbi = 0.0
        do 43 j=1,3
          dbi = dbi + demat(i,j)*bvect(j)
          zmatij = 0.0
          do 42 k=1,3
            zmatij = zmatij + demat(i,k)*dbds(k,j)
   42     continue
          zmat(i,j) = clam*zmatij
   43   continue
        dbvec(i) = dbi
        zmat(i,i) = zmat(i,i) + 1.0
   44 continue
      call invert(3,zmat,zinv)
c
cc..evaluate function
      do 45 i=1,3
        psifn(i) = gpsig(i) - gpsig0(i) + clam*dbvec(i)
   45 continue
c
cc..now form dlambda
      top = 0.0
      botm = 0.0
      do 48 i=1,3
        topi = 0.0
        botmi = 0.0
        do 47 j=1,3
          topi = topi + zinv(i,j)*psifn(j)
          botmi = botmi + zinv(i,j)*dbvec(j)
   47   continue
        top = top + avect(i)*topi
        botm = botm + avect(i)*botmi
   48 continue
      top = fvalu - top
      dlam = top/botm
cc..and hence form dsigma
      do 50 i=1,3
        dsi = 0.0
        do 49 j=1,3
          dsi = dsi + zinv(i,j)*(psifn(j) + dlam*dbvec(j))
   49   continue
        dsigma(i) = -dsi
   50 continue
cc..update lambda and gpsig
      clam = clam + dlam
      do 52 i=1,3
        gpsig(i) = gpsig(i) + dsigma(i)
   52 continue
cc..test for convergence - F(gpsig) small enough?
      call norm1(3,3,dsigma,dsnorm)
      call ycrit(triax,sigc,gpsig,fvalu)
      if (abs(fvalu).gt.ftoler) goto 40
      if (dsnorm.gt.stoler) goto 40
c
cc..update stress
   70   do 72 i=1,3
          strsg1(i,ij,l) = gpsig(i)
   72   continue
c
cc..end of stress iteration for this Gauss point
   75 continue
c
cc..end of element loop
   80 continue
c
      return
      end
c
c
c
        subroutine residu(melem,mpoin,mnode,mtotv,nelem,npoin
     1       ,ntotv,lnods,lnodn,coord,nkode,accld,strsg,resld)
c
      dimension lnods(melem,mnode),encod(2,8),nkode(mpoin),
     1         bmatx(3,16),gpwts(4),lnodn(melem),
     2         coord(mpoin,2),strsg(3,4,melem),resld(mtotv),
     3         shape(8),gploc(2,4),deriv(2,8),cartd(2,8),
     4         gpsig(3),accld(mtotv),enres(16)
c
cc..calculate residual (out-of-balance) force vector in resld
c
      do 5 i=1,ntotv
        resld(i) = 0.0
    5 continue
c
c...start loop over all elements
      do 80 l=1,nelem
c
      thick = 1.0
c...put nodal coords in encod
      nnode = lnodn(l)
      do 15 n=1,nnode
      nn = lnods(l,n)
      encod(1,n) = coord(nn,1)
      encod(2,n) = coord(nn,2)
   15 continue
c
c...put local coords of 2x2 gauss-pts in gploc
      call gauss(gploc,gpwts,ngaus)
c
c...collect element contribution to residual in enres
      do 20 i=1,16
        enres(i) = 0.0
   20 continue
c
c...start loop through gauss-pts
      do 75 ij=1,ngaus
c...put local coords of gauss-pt in s,t
      s = gploc(1,ij)
      t = gploc(2,ij)
c...get shape fns (shape) & derivs (deriv) in local coords
      call sfr(s,t,shape,deriv)
c...get xy-derivs (cartd) & derterminant of Jacobian (djacb)
      call jacob(encod,shape,deriv,cartd,djacb)
c...get B matrix 
      call bmat(cartd,bmatx)
c...store Gauss point stresses in gpsig
      do 25 i=1,3
        gpsig(i) = strsg(i,ij,l)
   25 continue
c
cc..add GP contribution to integral(B^T * sigma) to enres
      dvolu = thick*djacb*gpwts(ij)
      do 33 i=1,16
        gi = 0.0
        do 32 j=1,3
          gi = gi + bmatx(j,i)*gpsig(j)
   32   continue
        enres(i) = enres(i) + gi*dvolu
   33 continue
c
cc..end of Gauss point loop
   75 continue
c
cc..now subtract enres into the global residual vector
      do 77 n=1,nnode
        nn = lnods(l,n)
        ni = 2*(n-1)
        nni = 2*(nn-1)
        resld(nni+1) = resld(nni+1) - enres(ni+1)
        resld(nni+2) = resld(nni+2) - enres(ni+2)
   77 continue
c
cc..end of element loop
   80 continue
c
c...finally, add the current external applied loads
      do 85 i=1,ntotv
        resld(i) = resld(i) + accld(i)
   85 continue
c...and zero the components corresponding to fixed d.o.f.s
      do 90 np=1,npoin
        kode = nkode(np)
        npi = 2*(np-1)
c...separate kode out into two digits
        kx = kode/10
        ky = kode - 10*kx
        if (kx.eq.1) resld(npi+1) = 0.0
        if (ky.eq.1) resld(npi+2) = 0.0
   90 continue
c
      return
      end
c
c
c
      subroutine ycrit(triax,sigc,gpsig,fvalu)
      dimension gpsig(3)
cc..test stresses for plastic yield
c
cc..get principal stresses
      call princ(gpsig,sig1,sig3)
c
cc..Mohr-Coulomb yield function
      fvalu = sig1 - triax*sig3 - sigc
c
      return
      end
c
c
      subroutine princ(gpsig,sig1,sig3)
      dimension gpsig(3)
cc..form principal stresses from gpsig
c
      sx = gpsig(1)
      sy = gpsig(2)
      tau = gpsig(3)
      smean = 0.5*(sx + sy)
      sdev = 0.5*(sx - sy)
      sroot = sqrt(sdev*sdev + tau*tau)
      sig1 = smean + sroot
      sig3 = smean - sroot
c
      return
      end
c
c
c
      subroutine flowv(triax,gpsig,avect)
      dimension gpsig(3),avect(3),s1grad(3),s3grad(3)
cc..form flow vector (Mohr-Coulomb)
c
      sx = gpsig(1)
      sy = gpsig(2)
      tau = gpsig(3)
      call princ(gpsig,s1,s3)
      p = 0.5*(1 + (sx-sy)/(s1-s3))
      r = 2.0*tau/(s1-s3)
      s1grad(1) = p
      s1grad(2) = 1. - p
      s1grad(3) = r
      s3grad(1) = 1. - p
      s3grad(2) = p
      s3grad(3) = -r
      do 5 i=1,3
        avect(i) = s1grad(i) - triax*s3grad(i)
    5 continue
c
      return
      end
c
c
      subroutine flowm(triax,gpsig,dbds)
      dimension gpsig(3),dbds(3,3),vec(3)
c
cc..forms Jacobian of flow vector in dbds
      sx = gpsig(1)
      sy = gpsig(2)
      tau = gpsig(3)
      sd = 0.5*(sx - sy)
      sdt2 = sd*sd + tau*tau
      sroot = sqrt(sdt2)
      cc = (1.+triax)*0.25/(sdt2*sroot)
      vec(1) = -tau
      vec(2) = tau
      vec(3) = 2.*sd
      do 5 i=1,3
        do 5 j=1,3
          dbds(i,j) = cc*vec(i)*vec(j)
    5 continue
c
      return
      end
c
c
      subroutine invert(n,zmat,zinv)
      dimension zmat(n,n),zinv(n,n),zz(6,6)
c
cc..inverts zmat to get zinv
      do 10 i=1,3
        do 5 j=1,3
          zz(i,j) = zmat(i,j)
          zinv(i,j) = 0.0
    5   continue
        zinv(i,i) = 1.0
   10 continue
c
      do 35 i=1,n
        d = zz(i,i)
        k = i + 1
        if (k.gt.n) goto 40
        do 20 j=k,n
          c = zz(j,i)/d
          do 20 l=1,n
            zz(j,l) = zz(j,l) - zz(i,l)*c
            zinv(j,l) = zinv(j,l) - zinv(i,l)*c
   20   continue
   35 continue
   40 continue
c
      d = zz(n,n)
      do 42 j=1,n
        zinv(n,j) = zinv(n,j)/d
   42 continue
      do 50 k=2,n
        i = n - k + 1
        d = zz(i,i)
        l = i + 1
        do 48 jj=1,n
          t = zinv(i,jj)
          do 45 j=l,n
            t = t - zz(i,j)*zinv(j,jj)
   45     continue
          zinv(i,jj) = t/d
   48   continue
   50 continue
c
      return
      end
c
c
      subroutine check(mtotv,melem,mpoin,mnode,
     1ntotv,nelem,npoin,nkode,lnods,lnodn,
     1asdis,aslod,resld,rnorm,nminvc,lpoint)
      dimension asdis(mtotv),aslod(mtotv)
     1,resld(mtotv),nkode(npoin),nminvc(mtotv),lpoint(mtotv)
     2,lnods(melem,mnode),lnodn(melem)
c
c...find residual as gstif*asdis - aslod
c...use the element matrices, rather than gstif
c   
      call ematgv(18,mtotv,melem,mpoin,mnode,ntotv,nelem,
     1         npoin,lnods,lnodn,asdis,resld)
c      call gmult(mtotv,mstif,ntotv,npoin,nkode,
c     1gstif,asdis,resld,nminvc,lpoint)
c                     
c...finally, subtract aslod
      do 50 i=1,ntotv
        resld(i) = resld(i) - aslod(i)
   50 continue
c
c...zero the fixed d.o.f.s
      do 70 np=1,npoin
        kode = nkode(np)
        k1 = kode/10
        k2 = kode - k1*10
        nn = 2*(np-1)
        if (k1.eq.1) resld(nn+1) = 0
        if (k2.eq.1) resld(nn+2) = 0
   70 continue
c
c...now find the max. norm of resld
      call norm1(mtotv,ntotv,resld,rnorm)
c
      return
      end
c
c
      subroutine gmult(mtotv,mstif,ntotv,npoin,nkode,
     1gstif,asvin,asvout,nminvc,lpoint)
      dimension gstif(mstif),asvin(mtotv),asvout(mtotv)
     1,nkode(npoin),nminvc(mtotv),lpoint(mtotv)
c
c...multiply gstif*asvin and put product in asvout
c...first form L.asdis, where L is the lower triangle of gstif
c...this is done as in forward substitution in chosub
      asvout(1) = asvin(1)*gstif(1)
      do 10 i=2,ntotv
        sum = 0.0
        lpi = lpoint(i)
        lpi1 = lpoint(i-1)
        kstart = i - lpi + lpi1 + 1
c...note we end at k=i, not k=i-1
        do 5 k=kstart,i
          ikpos = gpos(lpoint,ntotv,i,k)
          sum = sum + gstif(ikpos)*asvin(k)
    5   continue
        asvout(i) = sum
   10 continue                                
c                   
c...now form y = U.x where U is upper triangle of gstif 
c...U does not include the diagonal elements of gstif
c...works like back-substitution in s/r chosub
      do 20 i=ntotv,2,-1
        lpi = lpoint(i)
        lpi1 = lpoint(i-1)
        ni = lpi - lpi1 - 1
        ai = asvin(i)
        do 15 k=1,ni
          clij = gstif(lpi-k)
          imk = i - k
          asvout(imk) = asvout(imk) + clij*ai
   15   continue
   20 continue
c
      return
      end
c
      subroutine norm1(m,n,v,vnorm)
      dimension v(m)
c
c...finds the max. norm of vector v
      vnorm = 0.0
      do 5 i=1,n
        vi = abs(v(i))
        if (vi.gt.vnorm) vnorm = vi
    5 continue
c
      return
      end
c
c...subroutines for conjugate gradient solution
c
      subroutine pcgsol(mtotv,melem,mpoin,mnode,ntotv,nelem,
     1 npoin,lnods,lnodn,diagk,nkode,mstif,gstif,nminvc,lpoint,
     2 asvin,asvout, z , r , p , y ,astmp,
     3 cgtol,itscg,rnorm0,isola,factor)
c
      dimension nkode(mpoin),diagk(mtotv),lnods(melem,mnode),
     1   lnodn(melem),kd(3),astmp(mtotv),
     2   asvout(mtotv),asvin(mtotv),z(mtotv),r(mtotv),p(mtotv),
     3   y(mtotv),gstif(mstif),nminvc(mtotv),lpoint(mtotv)
c
c...solve Bv=g using diagk as preconditioner
c
c...initial guess; 
      do 2 i=1,ntotv
c...we don't initialise asvout - use the previous solution
c...as first guess for this iteration
       asvout(i) = 0.0
       r(i) = asvin(i)
       z(i) = 0.0
    2 continue
c
c...zero the fixed d.o.f.s in r
      do 12 n=1,npoin
        kode = nkode(n)
        kd(1) = kode/10
        kd(2) = kode - kd(1)*10
        do 10 m=1,2
          if (kd(m).eq.0) goto 10
          nm = 2*(n - 1) + m
          r(nm) = 0.0
   10   continue
   12 continue
c
      if (isola.eq.3) then
c...incomplete Choleski decomposition of gstif
        call fixdof(mtotv,mstif,ntotv,mpoin,npoin,gstif,
     1            nkode,lpoint)
        call inchol(mtotv,mstif,ntotv,gstif,
     1                  lpoint,factor)
      endif
c
c...z = {C}^-1 * r, where C is the preconditioner
      if (isola.eq.2) then
c...diagonal precond., just divide by diagonal terms of K
        do 15 i=1,ntotv
          z(i) = r(i)/diagk(i)
   15   continue
      else
c...Incomplete Choleski preconditioner
        call chosub(mtotv,mstif,ntotv,gstif,
     1                r , z ,astmp,lpoint)
      endif
c
      do 16 i=1,ntotv
        p(i) = z(i)
   16 continue
      ztr = 0.0
      do 17 i=1,ntotv
        zi = z(i)
        ri = r(i)
        ztr = ztr + zi*ri
   17 continue
      call norm2(mtotv,ntotv,r,rnorm0)
      itscg = 0
c
      write(6,*)'PGSOL: iteration',itscg,',norm = ',rnorm0
c
   25 itscg = itscg + 1
c...create y which is K*p
      call ematgv(18,mtotv,melem,mpoin,mnode,ntotv,nelem,
     1         npoin,lnods,lnodn, p , y )
c
c...zero the fixed d.o.f.s in y
      do 26 n=1,npoin
        kode = nkode(n)
        kd(1) = kode/10
        kd(2) = kode - kd(1)*10
        do 24 m=1,2
          if (kd(m).eq.0) goto 24
          nm = 2*(n - 1) + m
          y(nm) = 0.0
   24   continue
   26 continue
c
    
      pty = 0.0
      do 28 i=1,ntotv
        pty = pty + p(i)*y(i)
   28 continue
      alpha = ztr/pty  
      do 30 i=1,ntotv
       asvout(i) = asvout(i) + alpha*p(i)
   30 continue
c
      do 40 i=1,ntotv
        r(i) = r(i) - alpha*y(i)
   40 continue
c
      call norm2(mtotv,ntotv,r,rnorm)
c
      write(6,*)'       iteration',itscg,',norm = ',rnorm
c
      if(rnorm.lt.cgtol) goto 200
c
c...z = {P}^-1 * r, where P is the preconditioner 
      if (isola.eq.2) then
        do 45 i=1,ntotv
          z(i) = r(i)/diagk(i)
   45   continue
      else
        call chosub(mtotv,mstif,ntotv,gstif,
     1                r , z ,astmp,lpoint)
      endif
c
      ztr0 = ztr
      ztr = 0.0
      do 60 i=1,ntotv
        ztr = ztr + z(i)*r(i)
   60 continue
      beta = ztr/ztr0
c
      do 130 i=1,ntotv
       p(i) = z(i) + beta*p(i)
  130 continue
c
      goto 25
c
  200 continue
      write(6,*)'Solution converged after ',itscg,' iterations.'
      write(6,*)
      return
      end
c
c
      subroutine norm2(mdim,ndim,v,vnorm)
      dimension v(mdim)
c
c...calculates the Euclidean norm of the vector v, returns it in vnorm
      sum = 0.0
      do 5 i=1,ndim
        vi = v(i)
        sum = sum + vi*vi
    5 continue
      vnorm = sqrt(sum)
      return
      end
c
c
      subroutine ematgv(ichan,mtotv,melem,mpoin,mnode,ntotv,nelem,
     1         npoin,lnods,lnodn,asvin,asvout)
      dimension asvin(mtotv),asvout(mtotv),estif(16,16),lgdof(16)
     1     ,lnods(melem,mnode),lnodn(melem)
c
c...multiply element matrices on file, 
c...by global vector asvin, to get global vector asvout
      rewind ichan
c
      do 10 i=1,ntotv
        asvout(i) = 0.0
   10 continue
c
c...loop through all elements
      do 50 l=1,nelem
        nnode = lnodn(l)
c...set up an array to relate the global node no.s to the local ones
        do 15 n=1,nnode
          np = lnods(l,n)
          nvarnp = 2*(np-1)
          do 12 i=1,2
            idof = 2*(n-1) + i
            iidof = nvarnp + i
            lgdof(idof) = iidof
   12     continue
   15   continue
        nevab = 2*nnode
c...the i'th d.o.f. of elt matrix corresponds to global d.o.f. lgdof(i)
c...note that this will not work in POREL.FOR, where an elt matrix may not
c...involve all d.o.f.s of a node (interface of structural and soil elts)
c 
        read (ichan) estif
c
        do 27 i=1,nevab
          sum = 0.0
          do 26 j=1,nevab
            jj = lgdof(j)
            sum = sum + estif(i,j)*asvin(jj)
   26     continue
          ii = lgdof(i)
          asvout(ii) = asvout(ii) + sum
   27   continue
c
   50 continue
c...end of element loop
c
      return
      end
c
c
      subroutine inchol(mtotv,mstif,ntotv,gstif,
     1                  lpoint,factor)
      dimension gstif(mstif),lpoint(mtotv)
c...Incomplete Choleski decomposition, using skyline storage 
c
      small = 1.e-8
c...count the amount of fill-in which occurs
      kfill = 0
      khole = 0
c...find the first diagonal entry, i.e. l(1,1)
      i = 1
      a11 = gstif(1)
      if (a11.lt.small) then
        write(6,*)'ERROR: zero or negative pivot in row ',i
        call prompt
        stop
      endif
      cl11 = sqrt(a11)
      gstif(1) = cl11
c...decompose each row in turn
      do 20 i=2,ntotv
        lpi = lpoint(i)
        lpi1 = lpoint(i-1)
        aii = gstif(lpi)
c...we will not fill in if the term is smaller than filtol
        filtol = factor*abs(aii)
        jstart = i - lpi + lpi1 + 1
c...work along the row
        do 10 j=jstart,i-1
          jjpos = lpoint(j)
          cljj = gstif(jjpos)
          ijpos = gpos(lpoint,ntotv,i,j)
          aij = gstif(ijpos)
          sum = 0.0
          lpj = lpoint(j)
          if (j.eq.1) then
            lpj1 = 0
          else
            lpj1 = lpoint(j-1)
          endif
          kstart=max(i-lpi+lpi1+1,j-lpj+lpj1+1,1)
          do 5 k=kstart,j-1
            ikpos = gpos(lpoint,ntotv,i,k)
            jkpos = gpos(lpoint,ntotv,j,k)
            if (ikpos.eq.0.or.jkpos.eq.0) goto 5
            sum = sum + gstif(ikpos)*gstif(jkpos)
    5     continue
          clij = (aij - sum)/cljj
c...here is the 'Incomplete' part
          ijpos = gpos(lpoint,ntotv,i,j)
          gij = gstif(ijpos)
          if (gij.eq.0.0) then
            khole = khole + 1
            if (abs(clij).lt.filtol) then
c...avoid fill-in: add i,j term to diagonal instead           
              gstif(lpi) = gstif(lpi) + clij
              goto 10
            endif
c...fill-in will occur
            kfill = kfill + 1
          endif         
          gstif(ijpos) = clij
   10   continue
c...finally find the diagonal entry on the row
        iipos = lpoint(i)
        aii = gstif(iipos)
        sum = 0.0
        kstart = i - lpi +lpi1 + 1
        do 15 k=kstart,i-1
          ikpos = gpos(lpoint,ntotv,i,k)
          clik = gstif(ikpos)
          if (clik.eq.0.0) goto 15
          sum = sum + clik*clik
   15   continue
        clii2 = aii - sum
        if (clii2.lt.small) then
          write(6,*)'ERROR: zero or negative pivot in row ',i
          call prompt
          stop
        endif
        clii = sqrt(clii2)
        iipos = lpoint(i)
        gstif(iipos) = clii
   20 continue
c
      write(6,*) 'No. of holes found in G matrix = ',khole
      write(6,*) 'No. of holes filled-in = ',kfill
c
      return
      end
c
c
      subroutine assdk(melem,mpoin,mnode,mtotv,
     1    nelem,npoin,lnods,lnodn,nkode,diagk)
c
c...assembles the diagonal of the stiffness matrices, 
c...read from the scratch file,
c...into the global vector diagk
c
      dimension lnods(melem,mnode),estif(16,16),kd(3),
     1         lnodn(melem),nkode(mpoin),diagk(mtotv)
c
c...initialize diagk
      do 20 i=1,mtotv
        diagk(i) = 0.0
   20 continue
c
      rewind 18
c
c...loop through all elements
      do 50 l=1,nelem
         read (18) estif
         nnode = lnodn(l)
c...deal with elt nodes ni in turn
c... these will be global nodes nni
        ni=0
        ndi = 0
   25   ni = ni + 1
        nni = lnods(l,ni)
        nvarnn = 2*(nni-1)
c...put the entries from estif into diag(gstif)
        do 40 mi=1,2
            im = ndi + mi
            iim = nvarnn + mi
            diagk(iim) = diagk(iim) + estif(im,im)
   40   continue
        ndi = ndi + 2
        if (ni.lt.nnode) goto 25
   50 continue
c...end of element loop
c
c...add a large number to the diagonal coefft of fixed d.o.f.s
      do 70 nn=1,npoin
        kode = nkode(nn)
        nvarnn = 2*(nn-1)
          kd(1) = kode/10
          kd(2) = kode - kd(1)*10
        do 60 m=1,2
          if (kd(m).eq.0) goto 60
          nposn = nvarnn + m
          diagk(nposn) = diagk(nposn) + 10.0e30
   60   continue
   70 continue
c
      return
      end
c
c
